Library Magvar;

uses
  SysUtils, Dialogs;

Type

  Param = Record
    OK: Char;
    Variation: Real;
    B_r: Real;
    B_theta: Real;
    B_phi: Real;
    B_x: Real;
    B_y: Real;
    B_z: Real;
  end;

  XFile = Array[0..255] of Char;
  XDate = Array[0..10] of Char;

var
  ModelFile: String;
  Buffer: String;
  MFile: TextFile;
  GnmModel: Array[0..12] of Array[0..12] of double;
  HnmModel: Array[0..12] of Array[0..12] of double;
  GtnmModel: Array[0..12] of Array[0..12] of double;
  HtnmModel: Array[0..12] of Array[0..12] of double;
  LastModel: String;
  Error: Boolean;

Const
  PiDiv = 0.01745329251994322958; {Pi / 180}
  Zero = 0.00000001;              {Virtual zero}

procedure AppException(Sender: TObject; E: Exception);
begin
  Error := True;
end;

procedure LoadModel;
var
  I, I2: Integer;
  S: String[8];
  Ds: String[1];
begin
  AssignFile(MFile, ModelFile);
  Reset(MFile);

  for I := 0 to 12 do begin {Load Gnm constants}
    Readln(MFile, Buffer);

    {Modifies for the correct Windows decimal separator}
    Ds := Buffer[7];
    if DecimalSeparator <> Ds then
      while Pos(Ds, Buffer) > 0 do
        Buffer[Pos(Ds, Buffer)] := DecimalSeparator;

    For I2 := 0 to 12 do begin
      S := Copy(Buffer, I2 * 8 + 1, 8);
      GnmModel[I][I2] := StrToFloat(S);
    end;
  end;

  for I := 0 to 12 do begin {Load Hnm constants}
    Readln(MFile, Buffer);

    {Modifies for the correct Windows decimal separator}
    if DecimalSeparator <> Ds then
      while Pos(Ds, Buffer) > 0 do
        Buffer[Pos(Ds, Buffer)] := DecimalSeparator;

    For I2 := 0 to 12 do begin
      S := Copy(Buffer, I2 * 8 + 1, 8);
      HnmModel[I][I2] := StrToFloat(S);
    end;
  end;

  for I := 0 to 12 do begin {Load Gtnm constants}
    Readln(MFile, Buffer);

    {Modifies for the correct Windows decimal separator}
    if DecimalSeparator <> Ds then
      while Pos(Ds, Buffer) > 0 do
        Buffer[Pos(Ds, Buffer)] := DecimalSeparator;

    For I2 := 0 to 12 do begin
      S := Copy(Buffer, I2 * 8 + 1, 8);
      GtnmModel[I][I2] := StrToFloat(S);
    end;
  end;

  for I := 0 to 12 do begin {Load Htnm constants}
    Readln(MFile, Buffer);

    {Modifies for the correct Windows decimal separator}
    if DecimalSeparator <> Ds then
      while Pos(Ds, Buffer) > 0 do
        Buffer[Pos(Ds, Buffer)] := DecimalSeparator;

    For I2 := 0 to 12 do begin
      S := Copy(Buffer, I2 * 8 + 1, 8);
      HtnmModel[I][I2] := StrToFloat(S);
    end;
  end;

  CloseFile(MFile);
  LastModel := ModelFile;
end;

Function MagVarC(XDat: XDate; Lat, Lon, h: Real; FileModel: XFile): Param; Export;
const
  a = 6378.16;                       {major radius (km) IAU66}
  f = 1.0 / 298.25;                  {Invers flattening IAU66}
  b = 6378.16 * (1.0 -1.0 / 298.25); {Minor radius b=a*(1-f)}
  r_0 = 6371.2;                      {"mean radius" for spherical harmonic expansion}

var
  Comod: String;
  LatR, LonR, Dat: double;
  ForX, ForY: Double;
  YearFrac, sr, r, c, s, theta, fn, fn_0, inv_s: double;
  SinLat: double;
  CosLat: double;
  n, m: Integer;
  mm, maxm: LongInt;
  c1_n, c2_n, c3_n, tmp: double;
  psi, sinpsi, cospsi: double;
  P: Array[0..12] of array[0..12] of double;
  DP: Array[0..12] of array[0..12] of double;
  gnm: Array[0..12] of array[0..12] of double;
  hnm: Array[0..12] of array[0..12] of double;
  sm: Array[0..12] of double;
  cm: Array[0..12] of double;
  root: array[0..12] of double;
  roots: Array[0..12] of array[0..12] of Array[0..1] of double;

begin
  Error := False;
  Result.OK := '9';  {Unknown error}
  ModelFile := StrPas(FileModel);

  if not FileExists(ModelFile) then begin
    Result.OK := '1'; {Model file does not exist}
    Exit;
  end;

  if ModelFile <> LastModel then {Load only if different from last used}
  try
    LoadModel;
  except
    Error := True;
    Result.OK := '2';  {Wrong or invalid model file}
    exit;
  end;

  Comod := ExtractFileName(ModelFile);
  Comod := Copy(Comod, Length(Comod) - 6, 4); {Year in Model file name}
  Dat := StrToDate(StrPas(XDat));
  try
    YearFrac := (Dat - (StrToDate('01' + DateSeparator + '01' + DateSeparator +
      Comod))) / 365.25;
  except
    Error := True;
    Result.OK := '3';  {Wrong or invalid data}
    exit;
  end;

  if Lat = 0 then Lat := Zero;
  if Lon = 0 then Lon := Zero;

  LatR := Lat * PiDiv; {convert to radiants}
  LonR := Lon * PiDiv;

  SinLat := Sin(LatR);
  if SinLat = 0 then SinLat := Zero;
  CosLat := Cos(LatR);
  if CosLat = 0 then CosLat := Zero;

  sr := Sqrt(a*a*CosLat*CosLat + b*b*SinLat*SinLat); {effective radius}
  ForX := CosLat * (h*sr + a*a);
  ForY := SinLat * (h*sr + b*b);
  Theta := ArcTan(ForX / ForY); {geocentric co-latitude}
  if (ForX >= 0) and (ForY < 0) then Theta := Theta + Pi;
  if (ForX < 0) and (ForY < 0) then Theta := Theta - Pi;
  r := h*h + 2.0*h * sr + (a*a*a*a - (a*a*a*a - b*b*b*b) * SinLat * SinLat) /
    (a*a - (a*a - b*b) * SinLat * SinLat);
  r := Sqrt(r); {geocentric radial distance}
  c := Cos(theta);
  s := Sin(theta);
  if s = 0 then inv_s := 1.0 / Zero
  else inv_s := 1.0 / s;

  {zero arrays}
  for n := 0 to 12 do
    for m := 0 to 12 do begin
      P[n][m] := 0;
      DP[n][m] := 0;
    end;

  {diagonal elements}
  P[0][0] := 1;
  P[1][1] := s;
  DP[0][0] := 0;
  DP[1][1] := c;
  P[1][0] := c;
  DP[1][0] := s * -1;

  for n := 2 to 12 do root[n] := sqrt((2.0*n-1) / (2.0*n));

  for m := 0 to 12 do begin
    mm := m*m;
    maxm := m + 1;
    if maxm < 2 then maxm := 2;
    for n := maxm to 12 do begin
      roots[m][n][0] := sqrt((n - 1) * (n - 1) - mm);
      roots[m][n][1] := 1.0 / sqrt(n * n - mm);
    end;
  end;

  for n := 2 to 12 do begin
    P[n][n] := P[n-1][n-1] * s * root[n];
    DP[n][n] := (DP[n-1][n-1] * s + P[n-1][n-1] * c) * root[n];
  end;

  {lower triangle}
  for m := 0 to 12 do begin
    maxm := m + 1;
    if maxm < 2 then maxm := 2;
    for n := maxm to 12 do begin
      P[n][m] := (P[n-1][m] * c * (2.0 * n-1) -
        P[n-2][m] * roots[m][n][0]) * roots[m][n][1];
      DP[n][m] := ((DP[n-1][m] * c - P[n-1][m] * s) *
        (2.0 * n-1) - DP[n-2][m] * roots[m][n][0]) * roots[m][n][1];
    end;
  end;

  {compute gnm, hnm at date}
  for n := 1 to 12 do
    for m := 0 to 12 do begin
      gnm[n][m] := gnmModel[n][m] + yearfrac * gtnmModel[n][m];
      hnm[n][m] := hnmModel[n][m] + yearfrac * htnmModel[n][m];
    end;

  {compute sm (sin(m lon) and cm (cos(m lon))}
  for m := 0 to 12 do begin
    sm[m] := sin(m * lonR);
    cm[m] := cos(m * lonR);
  end;

  {compute B fields}
  Result.B_r := 0;
  Result.B_Theta := 0;
  Result.B_Phi := 0;
  fn_0 := r_0 / r;
  fn := fn_0 * fn_0;

  for n := 1 to 12 do begin
    c1_n := 0;
    c2_n := 0;
    c3_n := 0;
    for m := 0 to n do begin
      tmp := (gnm[n][m] * cm[m] + hnm[n][m] * sm[m]);
      c1_n := c1_n + (tmp * P[n][m]);
      c2_n := c2_n + (tmp * DP[n][m]);
      c3_n := c3_n + (m * (gnm[n][m] * sm[m] - hnm[n][m] * cm[m]) * P[n][m]);
    end;
    { fn = pow(r_0/r,n+2)}
    fn := fn * fn_0;
    Result.B_r := Result.B_r + ((n + 1) * c1_n * fn);
    Result.B_theta := Result.B_theta - (c2_n * fn);
    Result.B_phi := Result.B_phi + (c3_n * fn * inv_s);
  end;

  {find geodetic field components}
  psi := theta - (PI / 2.0 - LatR);
  sinpsi := sin(psi);
  cospsi := cos(psi);
  Result.B_X := (Result.B_theta * -1) * cospsi - Result.B_r * sinpsi;
  Result.B_Y := Result.B_Phi;
  Result.B_Z := Result.B_Theta * sinpsi - Result.B_r * cospsi;
  if (Result.B_X = 0) or (Result.B_Y = 0) then Result.Variation := 0
  else begin
    Result.Variation := ArcTan(Result.B_Y / Result.B_X);
    if (Result.B_Y >= 0) and (Result.B_X < 0)
    then Result.Variation := Result.Variation + Pi;
    if (Result.B_X < 0) and (Result.B_Y < 0)
    then Result.Variation := Result.Variation - Pi;
    Result.Variation := Result.Variation / PiDiv; {convert to degrees}
  end;

  Result.OK := '0'; {all is OK}
end;

Exports
  MagVarC       index 1;

begin

end.
